@@ -12,6 +12,7 @@ import select
 import sys
 import tempfile
 import textwrap
+import traceback
 from enum import Enum
 from functools import partial
 from pathlib import Path
@@ -82,10 +83,10 @@ from . import RUNNABLE_KERNEL
 from . import RUNNABLE_SCRIPT
 from .exec import Argument
 from .exec import ArgumentKind
-from .exec import StatusType
 from .exec import Directory
 from .exec import FileName
 from .exec import FilePath
+from .exec import StatusType
 from .exec import get_arguments
 from .gui import IconLabel
 from .kernel import MyKernel
@@ -127,6 +128,49 @@ class HLine(QFrame):
         self.setFrameShape(QFrame.HLine | QFrame.Sunken)
 
 
+class RecurringTaskSignals(QObject):
+    """
+    Defines the signals available from a running recurring task thread.
+
+    Supported signals are:
+
+        finished
+            No data
+
+        error
+            tuple (exc_type, value, traceback.format_exc() )
+
+        result
+            object data returned from processing, anything
+
+    """
+    result = pyqtSignal(object)
+    finished = pyqtSignal()
+    error = pyqtSignal(tuple)
+
+
+class RecurringTask(QRunnable):
+    def __init__(self, func: Callable, *args, **kwargs):
+        super().__init__()
+        self._func = func
+        self._args = args
+        self._kwargs = kwargs
+        self.signals = RecurringTaskSignals()
+
+    @pyqtSlot()
+    def run(self):
+        try:
+            result = self._func(*self._args, **self._kwargs)
+        except Exception:
+            traceback.print_exc()
+            exc_type, value = sys.exc_info()[:2]
+            self.signals.error.emit((exc_type, value, traceback.format_exc()))
+        else:
+            self.signals.result.emit(result)  # Return the result of the processing
+        finally:
+            self.signals.finished.emit()  # Done
+
+
 class FunctionThreadSignals(QObject):
     """
     Defines the signals available from a running function thread.
@@ -442,7 +486,6 @@ class FunctionRunnableKernel(FunctionRunnable):
                 # self.signals.data.emit(Text.from_ansi('\n'.join(traceback)))
 
 
-
 class FunctionRunnableQProcess(FunctionRunnable):
     def __init__(self, func: Callable, args: List, kwargs: Dict, input_queue: Queue):
         super().__init__(func, args, kwargs, input_queue)
@@ -469,7 +512,7 @@ class FunctionRunnableQProcess(FunctionRunnable):
             # will be a: QProcess: Destroyed while process ("...venv/bin/python") is still running. .. after 30 seconds.
             self._process.waitForFinished(-1)
 
-        except (Exception,) as exc:
+        except Exception as exc:
             self.signals.error.emit(exc)
         finally:
             self.signals.finished.emit(self, self._func.__name__, True)
@@ -1093,7 +1136,9 @@ class View(QMainWindow):
         self._console_panel = ConsoleOutput()
 
         if len(self._buttons_panels) == 1:
-            self._buttons_widget = self._buttons_panels["Main"]
+            # If there is only one buttons panel, we do not create a TabWidget, but use that panel directly.
+            # We do not know the name (key) that was given in the returned dict, so we take the first item.
+            self._buttons_widget = list(self._buttons_panels.values())[0]
         else:
             self._buttons_widget = QTabWidget()
             for name, widget in self._buttons_panels.items():
@@ -1152,6 +1197,9 @@ class View(QMainWindow):
         self.kernel_panel = KernelPanel(self.kernel_name)
         self._toolbar.addWidget(self.kernel_panel)
 
+        self.threadpool = QThreadPool()
+        # print("Multithreading with maximum %d threads" % self.threadpool.maxThreadCount())
+
         self._timer = QTimer()
         self._timer.setInterval(1000)  # This interval shall be in the settings
         self._timer.timeout.connect(self.run_recurring_tasks)
@@ -1159,16 +1207,37 @@ class View(QMainWindow):
 
     def closeEvent(self, event: QCloseEvent) -> None:
         if self._kernel:
+            print("Shutting down Jupyter kernel.")
             self._kernel.shutdown()
+
         event.accept()
 
+        print("Waiting for recurring tasks to end.", end='', flush=True)
+        while not self.threadpool.waitForDone(100):
+            print(".", end='', flush=True)
+        print(flush=True)
+
+    def start_recurring_task(self, task: Callable):
+        # Pass the function to execute
+        worker = RecurringTask(task)  # Any other args, kwargs are passed to the run function
+        worker.signals.result.connect(partial(self.update_status, task))
+        worker.signals.finished.connect(self.end_recurring_task)
+
+        # Execute
+        self.threadpool.start(worker)
+
+    def end_recurring_task(self):
+        pass
+
     def run_recurring_tasks(self):
         for func in self._recurring_tasks:
-            response = func()
-            if func.__ui_status_type__ == StatusType.NORMAL:
-                self._status_bar.showMessage(response)
-            else:
-                self._status_bar_fixed_widget.setText(response)
+            self.start_recurring_task(func)
+
+    def update_status(self, func: Callable, msg: str):
+        if func.__ui_status_type__ == StatusType.NORMAL:
+            self._status_bar.showMessage(msg)
+        else:
+            self._status_bar_fixed_widget.setText(msg)
 
     def start_kernel(self, force: bool = False) -> MyKernel:
 
@@ -1282,32 +1351,34 @@ class View(QMainWindow):
         self.function_complete(func.__name__, True)
 
     def create_button_panels(self) -> Dict:
-        module_path = self._model.module_path
-
-        mod = importlib.import_module(module_path)
-        tab_order: List = getattr(mod, "UI_TAB_ORDER", None)
+        module_path_list: List = self._model.module_path_list
 
         buttons_panels = {}
 
-        # If we do not have sub packages, we will not create tabs, and we also only need one
-        # FunctionButtonsPanel which will be called "Main".
+        for module_path in module_path_list:
+            mod = importlib.import_module(module_path)
+            tab_order: List = getattr(mod, "UI_TAB_ORDER", None)
 
-        panel = FunctionButtonsPanel()
-        if self.add_buttons_to_panel(panel, module_path=module_path):
-            buttons_panels["Main"] = panel
+            # If we do not have sub packages, we will not create tabs, and we also only need one
+            # FunctionButtonsPanel which will be called "Main".
 
-        if subpackages := self._model.get_ui_subpackages():
-            if tab_order is None:
-                # Here we sort in display_name
-                sorted_subpackages = sorted(subpackages.items(), key=lambda x: x[1][0])
-            else:
-                # sorted_subpackages = sorted(subpackages.items(), key=lambda x: tab_order.index(x[0]))
-                # This way seems to be faster: see https://stackoverflow.com/a/21773891/4609203
-                sorted_subpackages = [(name, subpackages[name]) for name in tab_order if name in subpackages]
-            for name, (display_name, _) in sorted_subpackages:
-                panel = FunctionButtonsPanel()
-                self.add_buttons_to_panel(panel, module_path=f"{self._model.module_path}.{name}")
-                buttons_panels[display_name] = panel
+            panel = FunctionButtonsPanel()
+            if self.add_buttons_to_panel(panel, module_path=module_path):
+                tab_name = getattr(mod, "UI_TAB_DISPLAY_NAME", "Main")
+                buttons_panels[tab_name] = panel
+
+            if subpackages := self._model.get_ui_subpackages(module_path_list=[module_path]):
+                if tab_order is None:
+                    # Here we sort in display_name
+                    sorted_subpackages = sorted(subpackages.items(), key=lambda x: x[1][0])
+                else:
+                    # sorted_subpackages = sorted(subpackages.items(), key=lambda x: tab_order.index(x[0]))
+                    # This way seems to be faster: see https://stackoverflow.com/a/21773891/4609203
+                    sorted_subpackages = [(name, subpackages[name]) for name in tab_order if name in subpackages]
+                for name, (display_name, _) in sorted_subpackages:
+                    panel = FunctionButtonsPanel()
+                    self.add_buttons_to_panel(panel, module_path=f"{module_path}.{name}")
+                    buttons_panels[display_name] = panel
 
         return buttons_panels
 
@@ -1321,7 +1392,7 @@ class View(QMainWindow):
         Returns:
             The number of buttons added.
         """
-        modules = self._model.get_ui_modules(module_path=module_path)
+        modules = self._model.get_ui_modules(module_path_list=[module_path])
         number_of_buttons = 0
 
         for _, mod in sorted(modules.values()):
